# Copyright (c) 2018, 2020 Robert Clauscker <fuz@fuz.su>.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.

	.code16

	.set BUFSIZ,4096

	.bss
	.align 4
				# bp points to sum
sum:	.space 4		# current checksum
len:	.space 4		# current file length
fname:	.space 2		# file name
iobuf:	.space BUFSIZ

	.text
	.globl start
start:	cld

	# initialize data structures for checksum
	xor %ax,%ax
	mov $sum,%bp
	mov %bp,%di

	stosw			# clear sum
	stosw

	stosw			# clear len
	stosw

	# terminate argument vector
	mov $0x80,%di		# pointer to length of arguments
	xchg %ax, %bx		# clear bx
	mov (%di),%bl		# load argument vector length
	inc %di			# beginning of arguments
	movb %bh,(%bx,%di)	# NUL-terminate arguments

	# trim leading spaces
	mov $0x3d20, %ax	# AL = ' '
				# or AL = access mode DENYWRITE
				# for int 21/AH=3d OPEN EXISTING FILE
	mov %bx, %cx		# max length: argument vector length
	repe scasb		# find first non-space
	dec %di			# go back to first non-space
	mov %di,fname-sum(%bp)	# remember file name for later

	# open file
	mov %di,%dx		# file name
	int $0x21		# int 21/ah=3d open existing file
	push %ax		# remember handle
	jc 1f

0:	pop %bx
	push %bx		# load handle
	mov $BUFSIZ,%cx
	mov $iobuf,%dx
	mov $0x3f,%ah
	int $0x21		# int 21/ah=3f read
	jnc 2f			# error occured?

	pop %bx			# load handle
	push %ax		# remember error number for later
	mov $0x3e,%ah
	int $0x21		# int 21/ah=3e close file

1:	mov fname-sum(%bp),%si	# on failure print error message and exit
	call puts

	mov $colsp,%si
	call puts

	pop %bx			# on error, handle contains the error number
	shl %bx
	mov errors(%bx),%si	# look up the error message
	call puts

	mov $crlf,%si
	call puts

	mov $0x4c01,%ax		# exit with status 1: error
	int $0x21

2:	test %ax,%ax
	jz 1f
	add %ax,len-sum(%bp)	# add fill to len
	adcw $0,len+2-sum(%bp)
	xchg %ax,%cx		# check fill bytes
	call ckblk
	jmp 0b

1:	pop %bx			# load handle
	mov $0x3e,%ah
	int $0x21		# int 21/ah=3e close file

	# finalize check sum by appending length and inverting
	mov $len,%si
	mov $iobuf,%di
	movsw			# copy len
	movsw			# to iobuf

	mov $4,%bx

3:	cmp %bh,-5(%bx,%di)	# iobuf+bx-1 == 0?
	jnz 0f
	dec %bx
	jnz 3b

0:	mov %bx,%cx		# compute check sum over file length
	call ckblk

	mov %bp,%di
	notw (%di)		# sum+0
	notw 2(%di)		# sum+2

	call ltoa		# on di = sum

	mov $len,%di
	call ltoa

	mov fname-sum(%bp),%si
	call puts

	mov $crlf,%si
	jmp puts		# tail call, terminates the program

	# absorb CX bytes in iobuf into sum
ckblk:	mov (%bp),%ax		# sum+0
	mov 2(%bp),%dx		# sum+2
	mov $iobuf,%di

	jcxz 1f

	.align 4

0:	xor %bx, %bx
	mov (%di),%bl
	inc %di

	xor %dh,%bl
	shl %bx
	shl %bx

	mov %dl,%dh
	mov %ah,%dl
	mov %al,%ah
	xor %al,%al

	xor cktbl(%bx),%ax
	xor cktbl+2(%bx),%dx

	loop 0b

1:	mov %ax, (%bp)		# sum+0
	mov %dx, 2(%bp)		# sum+2
	ret

	# convert 32 bit number %di to asciz in iobuf
	# and then print it, followed by a space
	# high word in 2(%di)
	# low  word in 0(%di)
ltoa:	mov $iobuf+BUFSIZ-1,%si
	mov $10,%cx		# divisor
	movb %ch,(%si)		# NUL-terminate string

0:	xor %dx,%dx
	mov 2(%di),%ax
	div %cx			# pre-divide high word
	xchg %ax,%bx		# bx: new high word

	mov 0(%di),%ax
	div %cx			# actual division

	mov %ax,(%di)
	mov %bx,2(%di)
	add $'0',%dl
	dec %si
	mov %dl,(%si)

	test %bx,%bx		# are we done?
	jnz 0b
	test %ax,%ax
	jnz 0b

	call puts

	# print space
	mov $0x02,%ah
	mov $' ',%dl
	int $0x21		# int 21/ah=02 write character to standard output

1:	ret

	# print message in %si w/o newline

1:	xchg %ax,%dx
	mov $0x02,%ah
	int $0x21		# int 21/ah=02 write character to standard output

puts:	lodsb
	test %al,%al
	jnz 1b			# return if terminating NUL reached

	ret

	.data
colsp:	.string ": "
crlf:	.string "\r\n"
author:	.string "(c) 2018, 2020 Robert Clausecker <fuz@fuz.su>"

	.align 4
cktbl:	.int 0x00000000, 0x04c11db7, 0x09823b6e, 0x0d4326d9
	.int 0x130476dc, 0x17c56b6b, 0x1a864db2, 0x1e475005
	.int 0x2608edb8, 0x22c9f00f, 0x2f8ad6d6, 0x2b4bcb61
	.int 0x350c9b64, 0x31cd86d3, 0x3c8ea00a, 0x384fbdbd
	.int 0x4c11db70, 0x48d0c6c7, 0x4593e01e, 0x4152fda9
	.int 0x5f15adac, 0x5bd4b01b, 0x569796c2, 0x52568b75
	.int 0x6a1936c8, 0x6ed82b7f, 0x639b0da6, 0x675a1011
	.int 0x791d4014, 0x7ddc5da3, 0x709f7b7a, 0x745e66cd
	.int 0x9823b6e0, 0x9ce2ab57, 0x91a18d8e, 0x95609039
	.int 0x8b27c03c, 0x8fe6dd8b, 0x82a5fb52, 0x8664e6e5
	.int 0xbe2b5b58, 0xbaea46ef, 0xb7a96036, 0xb3687d81
	.int 0xad2f2d84, 0xa9ee3033, 0xa4ad16ea, 0xa06c0b5d
	.int 0xd4326d90, 0xd0f37027, 0xddb056fe, 0xd9714b49
	.int 0xc7361b4c, 0xc3f706fb, 0xceb42022, 0xca753d95
	.int 0xf23a8028, 0xf6fb9d9f, 0xfbb8bb46, 0xff79a6f1
	.int 0xe13ef6f4, 0xe5ffeb43, 0xe8bccd9a, 0xec7dd02d
	.int 0x34867077, 0x30476dc0, 0x3d044b19, 0x39c556ae
	.int 0x278206ab, 0x23431b1c, 0x2e003dc5, 0x2ac12072
	.int 0x128e9dcf, 0x164f8078, 0x1b0ca6a1, 0x1fcdbb16
	.int 0x018aeb13, 0x054bf6a4, 0x0808d07d, 0x0cc9cdca
	.int 0x7897ab07, 0x7c56b6b0, 0x71159069, 0x75d48dde
	.int 0x6b93dddb, 0x6f52c06c, 0x6211e6b5, 0x66d0fb02
	.int 0x5e9f46bf, 0x5a5e5b08, 0x571d7dd1, 0x53dc6066
	.int 0x4d9b3063, 0x495a2dd4, 0x44190b0d, 0x40d816ba
	.int 0xaca5c697, 0xa864db20, 0xa527fdf9, 0xa1e6e04e
	.int 0xbfa1b04b, 0xbb60adfc, 0xb6238b25, 0xb2e29692
	.int 0x8aad2b2f, 0x8e6c3698, 0x832f1041, 0x87ee0df6
	.int 0x99a95df3, 0x9d684044, 0x902b669d, 0x94ea7b2a
	.int 0xe0b41de7, 0xe4750050, 0xe9362689, 0xedf73b3e
	.int 0xf3b06b3b, 0xf771768c, 0xfa325055, 0xfef34de2
	.int 0xc6bcf05f, 0xc27dede8, 0xcf3ecb31, 0xcbffd686
	.int 0xd5b88683, 0xd1799b34, 0xdc3abded, 0xd8fba05a
	.int 0x690ce0ee, 0x6dcdfd59, 0x608edb80, 0x644fc637
	.int 0x7a089632, 0x7ec98b85, 0x738aad5c, 0x774bb0eb
	.int 0x4f040d56, 0x4bc510e1, 0x46863638, 0x42472b8f
	.int 0x5c007b8a, 0x58c1663d, 0x558240e4, 0x51435d53
	.int 0x251d3b9e, 0x21dc2629, 0x2c9f00f0, 0x285e1d47
	.int 0x36194d42, 0x32d850f5, 0x3f9b762c, 0x3b5a6b9b
	.int 0x0315d626, 0x07d4cb91, 0x0a97ed48, 0x0e56f0ff
	.int 0x1011a0fa, 0x14d0bd4d, 0x19939b94, 0x1d528623
	.int 0xf12f560e, 0xf5ee4bb9, 0xf8ad6d60, 0xfc6c70d7
	.int 0xe22b20d2, 0xe6ea3d65, 0xeba91bbc, 0xef68060b
	.int 0xd727bbb6, 0xd3e6a601, 0xdea580d8, 0xda649d6f
	.int 0xc423cd6a, 0xc0e2d0dd, 0xcda1f604, 0xc960ebb3
	.int 0xbd3e8d7e, 0xb9ff90c9, 0xb4bcb610, 0xb07daba7
	.int 0xae3afba2, 0xaafbe615, 0xa7b8c0cc, 0xa379dd7b
	.int 0x9b3660c6, 0x9ff77d71, 0x92b45ba8, 0x9675461f
	.int 0x8832161a, 0x8cf30bad, 0x81b02d74, 0x857130c3
	.int 0x5d8a9099, 0x594b8d2e, 0x5408abf7, 0x50c9b640
	.int 0x4e8ee645, 0x4a4ffbf2, 0x470cdd2b, 0x43cdc09c
	.int 0x7b827d21, 0x7f436096, 0x7200464f, 0x76c15bf8
	.int 0x68860bfd, 0x6c47164a, 0x61043093, 0x65c52d24
	.int 0x119b4be9, 0x155a565e, 0x18197087, 0x1cd86d30
	.int 0x029f3d35, 0x065e2082, 0x0b1d065b, 0x0fdc1bec
	.int 0x3793a651, 0x3352bbe6, 0x3e119d3f, 0x3ad08088
	.int 0x2497d08d, 0x2056cd3a, 0x2d15ebe3, 0x29d4f654
	.int 0xc5a92679, 0xc1683bce, 0xcc2b1d17, 0xc8ea00a0
	.int 0xd6ad50a5, 0xd26c4d12, 0xdf2f6bcb, 0xdbee767c
	.int 0xe3a1cbc1, 0xe760d676, 0xea23f0af, 0xeee2ed18
	.int 0xf0a5bd1d, 0xf464a0aa, 0xf9278673, 0xfde69bc4
	.int 0x89b8fd09, 0x8d79e0be, 0x803ac667, 0x84fbdbd0
	.int 0x9abc8bd5, 0x9e7d9662, 0x933eb0bb, 0x97ffad0c
	.int 0xafb010b1, 0xab710d06, 0xa6322bdf, 0xa2f33668
	.int 0xbcb4666d, 0xb8757bda, 0xb5365d03, 0xb1f740b4
	.size cktbl,.-cktbl

	.align 2
errors:	.short .LXX, .L01, .L02, .L03, .L04, .L05, .L06, .LXX
	.short .LXX, .LXX, .LXX, .LXX, .L0C, .LXX, .LXX, .LXX
	.short .LXX, .LXX, .LXX

.L01:	.string "function number invalid"	# 3D
.L02:	.string "file not found"		# 3D
.L03:	.string "path not found"		# 3D
.L04:	.string "too many open files"		# 3D
.L05:	.string "access denied"			# 3D 3F
.L06:	.string "invalid handle"		# 3E 3F
.L0C:	.string "access code invalid"		# 3D
.LXX:	.string "unknown error"			# catch all
